/*-------------------------------------------------------------------------+
| Copyright 2016 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.create;

import java.util.List;
import java.util.Set;

import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.decode.ComposableDecoder;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.exception.DecodingException;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.genotype.ComposableGenotype;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.genotype.FlatGenotypeMap;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.phenotype.FlatPhenotypeMap;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.phenotype.Phenotype;
import org.fortiss.af3.exploration.alg.exception.ExplorationException;
import org.fortiss.af3.exploration.alg.service.IExplorationEncoding;
import org.opt4j.core.Genotype;
import org.opt4j.core.genotype.CompositeGenotype;
import org.opt4j.core.problem.Creator;
import org.opt4j.core.problem.Decoder;

/**
 * Interface for creating {@link ComposableGenotype}s, i.e. {@link Genotype}s that compose a
 * {@link CompositeGenotype}.
 * 
 * @author diewald
 */
public interface ComposableCreator<G extends ComposableGenotype> {

	/**
	 * Creates a {@link ComposableGenotype}. This method is called by the parent
	 * {@link CompositeGenotype}.
	 * 
	 * @param createdGenotypes
	 *            Map that lists already created {@link Genotype}s by their Type.
	 * @param decodedPhenotypes
	 *            Map of already generated phenotypes.
	 * @return The {@link ComposableGenotype}. May be a {@link CompositeGenotype} itself.
	 * @throws ExplorationException
	 */
	G createComposable(FlatGenotypeMap createdGenotypes,
			FlatPhenotypeMap<Phenotype> decodedPhenotypes) throws ExplorationException;

	/**
	 * Registers {@link Decoder}s that decode the {@link Genotype} created by {@code this}
	 * {@link Creator}. It is absolutely required to launch these {@link Decoder}s after the
	 * creation such that the generated {@link Phenotype}s are available for other {@link Creator}s
	 * that require these {@link Phenotype}s and which are launched subsequently to {@code this}
	 * {@link Creator}.
	 * 
	 * @param decoders
	 *            List of {@link Decoder}s that are executed in the order of the List.
	 */
	void registerAssocDecoders(List<ComposableDecoder<G, ? extends Phenotype>> decoders);

	/**
	 * Registers {@link Decoder}s that decode the {@link Genotype} created by {@code this}
	 * {@link Creator}. It is absolutely required to launch these {@link Decoder}s after the
	 * creation such that the generated {@link Phenotype}s are available for other {@link Creator}s
	 * that require these {@link Phenotype}s and which are launched subsequently to {@code this}
	 * {@link Creator}.
	 * 
	 * @param decoders
	 *            {@link Decoder}s that are executed in the order of the parameter list.
	 */
	void registerAssocDecoders(@SuppressWarnings("unchecked") ComposableDecoder<G, ?>... decoders);

	/**
	 * Decodes the created {@link Phenotype} using the set of registered {@link Decoder}s. The
	 * generated {@link Phenotype}s are returned in the form of {@link FlatPhenotypeMap}. The given
	 * {@link FlatPhenotypeMap} encodes {@link Phenotype}s that were already decoded earlier in the
	 * exploration process.
	 * 
	 * @param genotype
	 *            Created {@link Genotype} that shall be decoded.
	 * @param decodedPhenotypes
	 *            {@link FlatPhenotypeMap} of already decoded {@link Phenotype}s.
	 * @param createdGenotypeTypeSet
	 *            TODO
	 * @return {@link FlatPhenotypeMap} of decoded {@link Phenotype} by the registered
	 *         {@link Decoder}s.
	 * @throws DecodingException
	 */
	<P extends Phenotype> FlatPhenotypeMap<Phenotype> decodeCreatedGenotype(G genotype,
			FlatPhenotypeMap<Phenotype> decodedPhenotypes,
			Set<Class<? extends Genotype>> createdGenotypeTypeSet) throws DecodingException;

	/**
	 * Launches an identity decoder if there is any registered for {@code this}
	 * {@link ComposableCreator}. An identity decoder is a decoder that injects a hybrid genotype in
	 * the {@link FlatPhenotypeMap} such that it can be accessed by other decoders.
	 * 
	 * @param genotype
	 *            The created {@link Genotype}.
	 * @param decodedPhenotypes
	 *            Previously decoded {@link Phenotype}s.
	 * @return Modified {@code decodedPhenotype} map.
	 * @throws DecodingException
	 *             if the identity decoder fails.
	 */
	<P extends Phenotype> FlatPhenotypeMap<Phenotype> identityDecodeCreatedGenotype(G genotype,
			FlatPhenotypeMap<Phenotype> decodedPhenotypes) throws DecodingException;

	/** Returns the type of the created {@link IExplorationEncoding} / {@link Genotype}. */
	Class<? extends IExplorationEncoding> getCreatedEncodingType();
}
