/*-------------------------------------------------------------------------+
| Copyright 2014 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.backend.opt4j.problem;

import java.util.Collection;
import java.util.HashSet;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;

import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.phenotype.Phenotype;
import org.fortiss.af3.exploration.model.ExplorationTarget;
import org.opt4j.core.Genotype;
import org.opt4j.core.Objective;
import org.opt4j.core.problem.Creator;
import org.opt4j.core.problem.Decoder;
import org.opt4j.core.problem.Evaluator;
import org.opt4j.core.problem.ProblemModule;

import com.google.inject.TypeLiteral;
import com.google.inject.multibindings.Multibinder;

/**
 * Base class for a problem module that enables the registration of pre-constructed
 * creators/decoders/evaluators, and which thus can be constructed using a non-default constructor.
 * 
 * @author barner
 */
public abstract class ProblemModuleBase<G extends Genotype, R extends Creator<G>, D extends Decoder<G, ?>>
		extends org.opt4j.core.problem.ProblemModule {

	/** Stores the {@link Creator} bound to this {@link ProblemModule}. */
	protected R creator;

	/** Stores the {@link Decoder} bound to this {@link ProblemModule}. */
	protected D decoder;

	/** Stores the {@link Evaluator}s bound to this {@link ProblemModule}. */
	protected Collection<EvaluatorWrapper<?, ?, ?>> evaluators;

	/** Map that stores the associations between {@link ExplorationTarget}s and Objectives. */
	protected Map<ExplorationTarget<?>, Objective> targetObjectiveAssoc;

	/** Constructor. */
	public ProblemModuleBase() {
		evaluators = new HashSet<>();
		targetObjectiveAssoc = new ConcurrentHashMap<>();
	}

	/** Returns the {@link Objective} associated with the given {@link ExplorationTarget}. */
	public Objective getObjectiveOf(ExplorationTarget<?> target) {
		return targetObjectiveAssoc.get(target);
	}

	/**
	 * Registers a pre-constructed evaluator instances with the Opt4J framework.
	 */
	protected void addEvaluator(EvaluatorWrapper<?, ?, ?> evaluator) {
		assert (false);
		registerEvaluator(evaluator);
	}

	/** Adds the given {@link Evaluator} to the Opt4J framework. */
	private void addEvaluator(Evaluator<?> evaluator) {
		// It is crucial that the Multibinder and the TypeLiteral are defined for
		// <Evaluator<Object>, since otherwise the binding of the evaluators will fail.
		Multibinder<Evaluator<Object>> multibinder =
				Multibinder.newSetBinder(binder(), new TypeLiteral<Evaluator<Object>>() { //
						});

		addEvaluator(evaluator, multibinder);
	}

	/**
	 * Registers the given {@link EvaluatorWrapper} with {@code this} class such that it can be used
	 * for the result validation.
	 */
	protected void registerEvaluator(EvaluatorWrapper<?, ?, ?> evaluator) {
		evaluators.add(evaluator);
	}

	/**
	 * Registers a {@link Collection} of pre-constructed evaluator instances with the Opt4J
	 * framework.
	 */
	protected void addEvaluators(Collection<EvaluatorWrapper<?, ?, ?>> evaluators) {

		// It is crucial that the Multibinder and the TypeLiteral are defined for
		// <Evaluator<Object>, since otherwise the binding of the evaluators will fail.
		Multibinder<Evaluator<Object>> multibinder =
				Multibinder.newSetBinder(binder(), new TypeLiteral<Evaluator<Object>>() { //
						});

		// Configure evaluators (based on provided goal specification)
		for(EvaluatorWrapper<?, ?, ?> eval : evaluators) {
			addEvaluator(eval, multibinder);
			evaluators.add(eval);
		}
	}

	/** Helper method to register a pre-constructed evaluator instance with the Opt4J framework. */
	@SuppressWarnings("unchecked")
	private void addEvaluator(Evaluator<? extends Object> evaluator,
			Multibinder<Evaluator<Object>> multibinder) {

		// It is crucial that the Multibinder and the TypeLiteral are defined for
		// <Evaluator<Object>, since otherwise the binding of the evaluators will fail.

		// This is inspired from ProblemModule.addEvaluator(), which does not allow to pass
		// parameters to the constructor of the evaluators. Hence, evaluators are instantiated here,
		// and the binding is performed against these instances.
		multibinder.addBinding().toInstance((Evaluator<Object>)evaluator);
	}

	// /** Registers a pre-constructor creator and decoder with the Opt4J engine. */
	// @SuppressWarnings("unchecked")
	// protected void bindProblem(R creator, D decoder) {
	// registerCreatorDecoder(creator, decoder);
	//
	// // This is inspired from ProblemModule.bindProblem(), which does not allow to pass
	// // parameters to the constructor of the creators/decoders. Hence, a pre-constructed
	// // creator/decoder is bound here.
	//
	// // Note that in contrast to ProblemModule.bindProblem(), the creator and the decoder
	// // are not bound in the singleton scope, since these classes need to be parameterized
	// // with the current input model.
	// binder().bind(new TypeLiteral<Creator<Genotype>>() { // empty
	// }).toInstance((Creator<Genotype>)creator);
	//
	// binder().bind(new TypeLiteral<Decoder<Genotype, Object>>() { // empty
	// }).toInstance((Decoder<Genotype, Object>)decoder);
	// }

	/** Registers a pre-constructor creator and decoder with the Opt4J engine. */
	@SuppressWarnings("unchecked")
	protected void bindProblem(Creator creator, Decoder decoder,
			Evaluator<? extends Phenotype> evaluator) {
		registerCreatorDecoder((R)creator, (D)decoder);

		// This is inspired from ProblemModule.bindProblem(), which does not allow to pass
		// parameters to the constructor of the creators/decoders. Hence, a pre-constructed
		// creator/decoder is bound here.

		// Note that in contrast to ProblemModule.bindProblem(), the creator and the decoder
		// are not bound in the singleton scope, since these classes need to be parameterized
		// with the current input model.
		binder().bind(new TypeLiteral<Creator<Genotype>>() { // empty
				}).toInstance((Creator<Genotype>)creator);

		// FIXME: Duplicates below!!!!!
		binder().bind(new TypeLiteral<Decoder<Genotype, Phenotype>>() { // empty
				}).toInstance((Decoder<Genotype, Phenotype>)decoder);

		binder().bind(new TypeLiteral<Decoder<Genotype, Object>>() { // empty
				}).toInstance((Decoder<Genotype, Object>)decoder);

		addEvaluator(evaluator);
	}

	// /** Registers a pre-constructor creator, decoder, and evaluator with the Opt4J engine. */
	// @SuppressWarnings("unchecked")
	// // TODO: remove casts and rethrow the exception.
	// protected
	// void bindProblem(R creator, D decoder, Evaluator<? extends Phenotype> evaluator) {
	// registerCreatorDecoder(creator, decoder);
	//
	// // This is inspired from ProblemModule.bindProblem(), which does not allow to pass
	// // parameters to the constructor of the creators/decoders. Hence, a pre-constructed
	// // creator/decoder is bound here.
	//
	// // Note that in contrast to ProblemModule.bindProblem(), the creator and the decoder
	// // are not bound in the singleton scope, since these classes need to be parameterized
	// // with the current input model.
	// binder().bind(new TypeLiteral<Creator<Genotype>>() { // empty
	// }).toInstance((Creator<Genotype>)creator);
	//
	// binder().bind(new TypeLiteral<Decoder<Genotype, Object>>() { // empty
	// }).toInstance((Decoder<Genotype, Object>)decoder);
	//
	// addEvaluator(evaluator);
	// }

	/**
	 * Stores the associated creator and decoder at this class, if they are not already set. (They
	 * might be set for this {@link ProblemModule} already, if multiple {@link Evaluator}s are
	 * defined).
	 */
	private void registerCreatorDecoder(R creator, D decoder) {
		if(this.creator == null) {
			this.creator = creator;
		}
		if(this.decoder == null) {
			this.decoder = decoder;
		}
	}

	/** Returns the set of evaluators bound to this instance of the {@link ProblemModule}. */
	public Collection<EvaluatorWrapper<?, ?, ?>> getEvaluators() {
		return evaluators;
	}
}
