/*-------------------------------------------------------------------------+
| Copyright 2015 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.sysmodel.arch.taskgraph;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.instantiatedmapping.InstantiatedTaskMappingEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.instantiatedmapping.InstantiatedTaskMappingEntry;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.ITaskAdapter;
import org.jgrapht.graph.DefaultListenableGraph;

/**
 * Manages the construction of {@link TaskGraph}s whose vertices are exclusively instances of
 * {@link ITaskAdapter}s.
 * 
 * @author diewald
 */
public class TaskGraphInstantiator<T extends TaskGraph> {

	/**
	 * Holds for each {@link TaskGraph} in the system a map that relates possibly abstract
	 * {@link ITaskAdapter} with the {@link ITaskAdapter} that
	 * instantiate them.
	 */
	private Map<T, Map<ITaskAdapter<?>, ITaskAdapter<?>>> instanceToAbstractNodeMap;

	/**
	 * Constructs a {@link TaskGraphInstantiator} that can create a {@link TaskGraph} containing
	 * only {@link ITaskAdapter} instances from the given {#link abstractGraph}
	 * using the given {#link encoding}.
	 */
	public TaskGraphInstantiator(T abstractGraph,
			InstantiatedTaskMappingEncoding<? extends InstantiatedTaskMappingEntry> encoding) {
		instanceToAbstractNodeMap = new HashMap<T, Map<ITaskAdapter<?>, ITaskAdapter<?>>>();
		instanceToAbstractNodeMap.put(abstractGraph,
				createAbstractInstantiatedNodeAssociations(encoding));
	}

	/**
	 * Constructs a {@link TaskGraphInstantiator} that can create {@link TaskGraph}s containing
	 * only {@link ITaskAdapter} instances from the given {#link abstractGraphs}
	 * using the given {#link encoding}.
	 */
	public TaskGraphInstantiator(Collection<T> abstractGraphs,
			InstantiatedTaskMappingEncoding<? extends InstantiatedTaskMappingEntry> encoding) {
		instanceToAbstractNodeMap = new HashMap<T, Map<ITaskAdapter<?>, ITaskAdapter<?>>>();
		for(T taskGraph : abstractGraphs) {
			instanceToAbstractNodeMap.put(taskGraph,
					createAbstractInstantiatedNodeAssociations(encoding));
		}
	}

	/**
	 * Creates {@link TaskGraph}s containing only {@link ITaskAdapter} instances for
	 * each {@link TaskGraph} in the given collection.
	 */
	public Collection<T> createMultipleInstanceTaskGraphs(Collection<T> abstractGraphs) {
		Collection<T> instanceTaskGraphs = new ArrayList<T>();
		for(T taskGraph : abstractGraphs) {
			instanceTaskGraphs.add(createInstanceTaskGraph(taskGraph));
		}
		return instanceTaskGraphs;
	}

	/**
	 * Creates a single {@link TaskGraph} containing only {@link ITaskAdapter}
	 * instances for the given {@link TaskGraph}.
	 */
	public T createInstanceTaskGraph(T abstractGraph) {
		T instanceTaskGraph = abstractGraph.copy();

		Map<ITaskAdapter<?>, ITaskAdapter<?>> nodeCorrespondenceMap =
				instanceToAbstractNodeMap.get(abstractGraph);
		assert (nodeCorrespondenceMap != null);

		Set<ITaskAdapter<?>> removeableVertices = new HashSet<ITaskAdapter<?>>();
		for(Entry<ITaskAdapter<?>, ITaskAdapter<?>> instancePair : nodeCorrespondenceMap
				.entrySet()) {
			ITaskAdapter<?> instComp = instancePair.getKey();
			ITaskAdapter<?> abstrComp = instancePair.getValue();
			if(abstractGraph.containsTask(abstrComp) && abstrComp.isInstantiationRequired()) {
				insertInstanceComponent(instanceTaskGraph, abstrComp, instComp);
				// Add the instantiated "abstract" Component to the set of Components which shall be
				// removed from the instanceGraph. Here, a set of these components is needed to not
				// trying to remove the same abstract component twice by accident.
				removeableVertices.add(abstrComp);
			}
		}
		for(ITaskAdapter<?> abstractComponent : removeableVertices) {
			instanceTaskGraph.removeTask(abstractComponent);
		}

		return instanceTaskGraph;
	}

	/**
	 * Replaces the given {#link abstractNode} with the given {#link instanceNode} in the {#link
	 * instanceTaskGraph}.
	 */
	public void insertInstanceComponent(T instanceTaskGraph, ITaskAdapter<?> abstractNode,
			ITaskAdapter<?> instanceNode) {
		DefaultListenableGraph<ITaskAdapter<?>, SignalAdapterWeightedEdge> graph =
				instanceTaskGraph.getGraph();
		instanceTaskGraph.addTask(instanceNode, abstractNode);

		Collection<SignalAdapterWeightedEdge> incomingEdges = graph.incomingEdgesOf(abstractNode);
		for(SignalAdapterWeightedEdge incoming : incomingEdges) {
			graph.addEdge(graph.getEdgeSource(incoming), instanceNode,
					new SignalAdapterWeightedEdge(incoming.getSignalAdapter()));
		}

		Collection<SignalAdapterWeightedEdge> outgoingEdges = graph.outgoingEdgesOf(abstractNode);
		for(SignalAdapterWeightedEdge outgoing : outgoingEdges) {
			graph.addEdge(instanceNode, graph.getEdgeTarget(outgoing),
					new SignalAdapterWeightedEdge(outgoing.getSignalAdapter()));
		}
	}

	/**
	 * Creates a map which relates "abstract" {@link IDeployableComponentAdapter} with the
	 * {@link IDeployableComponentAdapter} instantiating the "abstract"
	 * {@link IDeployableComponentAdapter}.
	 */
	/**
	 * Creates a map which relates "abstract" {@link ITaskAdapter} with the
	 * {@link ITaskAdapter} instantiating the "abstract"
	 * {@link ITaskAdapter}.
	 */
	private Map<ITaskAdapter<?>, ITaskAdapter<?>> createAbstractInstantiatedNodeAssociations(
			InstantiatedTaskMappingEncoding<? extends InstantiatedTaskMappingEntry> encoding) {
		Map<ITaskAdapter<?>, ITaskAdapter<?>> instanceToAbstractNodeMap =
				new HashMap<ITaskAdapter<?>, ITaskAdapter<?>>();
		for(ITaskAdapter<?> instantiatedComponent : encoding.getRequesters()) {
			if(instantiatedComponent.isInstantiation()) {
				instanceToAbstractNodeMap.put(instantiatedComponent,
						instantiatedComponent.getReplacedComponent());
			} else {
				instanceToAbstractNodeMap.put(instantiatedComponent, instantiatedComponent);
			}
		}

		// assert (getMissingComponentInstantiations(abstractGraph).isEmpty()) :
		// "Some abstract Components are missing an instatiation in the encoding.";
		return instanceToAbstractNodeMap;
	}
}
