/*-------------------------------------------------------------------------+
| Copyright 2018 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.service.internal;

import java.util.Collection;
import java.util.HashSet;
import java.util.Set;
import java.util.stream.Collectors;

import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.phenotype.Phenotype;
import org.fortiss.af3.exploration.alg.dse.module.CreatorModule;
import org.fortiss.af3.exploration.alg.dse.module.DecoderModule;
import org.fortiss.af3.exploration.alg.dse.module.EvaluatorModule;
import org.fortiss.af3.exploration.alg.dse.module.OperatorModule;
import org.fortiss.af3.exploration.alg.exception.ExplorationServiceRuntimeException;
import org.fortiss.af3.exploration.alg.service.IExplorationModuleService;
import org.fortiss.tooling.kernel.introspection.IIntrospectionDetailsItem;
import org.fortiss.tooling.kernel.introspection.IIntrospectionItem;
import org.opt4j.core.Genotype;
import org.opt4j.core.problem.Decoder;

import com.google.common.collect.ClassToInstanceMap;
import com.google.common.collect.MutableClassToInstanceMap;
import com.google.inject.AbstractModule;

/**
 * Implementation of the {@link IExplorationModuleService}.
 * 
 * @author diewald
 */
public class ExplorationModuleService implements IExplorationModuleService {

	/** Set referencing all registered optimization variables. */
	private Set<Class<? extends Genotype>> genotypes = new HashSet<>();

	/** Set referencing all register candidate solution types. */
	private Set<Class<? extends Phenotype>> phenotypes = new HashSet<>();

	/**
	 * References all registered exploration modules. Differing between {@link Genotype}s,
	 * {@link Decoder}s etc. is currently not needed.
	 */
	private ClassToInstanceMap<AbstractModule> registeredExpModules = MutableClassToInstanceMap
			.create();

	/** Holds the instance of the evaluation service. */
	private volatile static ExplorationModuleService INSTANCE;

	/** Constructor. */
	private ExplorationModuleService() {
		// Prevent direct instantiation.
	}

	/** The instance (singleton) of the ExplorationTransformationService. */
	public static synchronized ExplorationModuleService getInstance() {
		if(INSTANCE == null) {
			INSTANCE = new ExplorationModuleService();
		}
		return INSTANCE;
	}

	/** {@inheritDoc} */
	@Override
	public void initializeService() {
		// Typically used to read extension point configs. Not needed since we won't use eclipse
		// extension points in the DSE any more.
	}

	/** {@inheritDoc} */
	@Override
	public void startService() {
		// Not needed.
	}

	/** {@inheritDoc} */
	@Override
	public <G extends Genotype> void registerVariable(Class<G> genotype, CreatorModule<G> creator,
			OperatorModule<G> operator) {
		// Input check: Allow only fully specified variable registrations.
		if(genotype == null) {
			throw new ExplorationServiceRuntimeException(getClass(), "The passed genotype is null!");
		} else if(creator == null) {
			throw new ExplorationServiceRuntimeException(getClass(), "The passed creator is null!");
		} else if(operator == null) {
			throw new ExplorationServiceRuntimeException(getClass(), "The passed operator is null!");
		}

		// Input check: Register modules only once.
		@SuppressWarnings("unchecked") Class<CreatorModule<G>> creatorClass =
				(Class<CreatorModule<G>>)creator.getClass();
		if(registeredExpModules.getInstance(creatorClass) != null) {
			throw new ExplorationServiceRuntimeException(getClass(), "The module " +
					creatorClass.getSimpleName() + " has already been registered!");
		}
		@SuppressWarnings("unchecked") Class<OperatorModule<G>> operatorClass =
				(Class<OperatorModule<G>>)operator.getClass();
		if(registeredExpModules.getInstance(operatorClass) != null) {
			throw new ExplorationServiceRuntimeException(getClass(), "The module " +
					operatorClass.getSimpleName() + " has already been registered!");
		}

		genotypes.add(genotype);
		registeredExpModules.putInstance(creatorClass, creator);
		registeredExpModules.putInstance(operatorClass, operator);
	}

	/** {@inheritDoc} */
	@Override
	public <P extends Phenotype> void registerProblem(Class<P> phenotype, DecoderModule<P> decoder) {
		// Input check: Allow only fully specified variable registrations.
		if(phenotype == null) {
			throw new ExplorationServiceRuntimeException(getClass(),
					"The passed phenotype is null!");
		} else if(decoder == null) {
			throw new ExplorationServiceRuntimeException(getClass(), "The passed decoder is null!");
		}

		// Input check: Register modules only once.
		@SuppressWarnings("unchecked") Class<DecoderModule<P>> decoderClass =
				(Class<DecoderModule<P>>)decoder.getClass();
		if(registeredExpModules.getInstance(decoderClass) != null) {
			throw new ExplorationServiceRuntimeException(getClass(), "The module " +
					decoderClass.getSimpleName() + " has already been registered!");
		}

		phenotypes.add(phenotype);
		registeredExpModules.putInstance(decoderClass, decoder);
	}

	/** {@inheritDoc} */
	@Override
	public void registerEvaluator(EvaluatorModule<Phenotype> evaluator) {
		// Input check: Allow only fully specified variable registrations.
		if(evaluator == null) {
			throw new ExplorationServiceRuntimeException(getClass(),
					"The passed evaluator is null!");
		}

		// Input check: Register modules only once.
		@SuppressWarnings("unchecked") Class<EvaluatorModule<Phenotype>> evaluatorClass =
				(Class<EvaluatorModule<Phenotype>>)evaluator.getClass();
		if(registeredExpModules.getInstance(evaluatorClass) != null) {
			throw new ExplorationServiceRuntimeException(getClass(), "The module " +
					evaluatorClass.getSimpleName() + " has already been registered!");
		}

		registeredExpModules.putInstance(evaluatorClass, evaluator);
	}

	/** {@inheritDoc} */
	@Override
	public Set<CreatorModule<?>> getCreators() {
		return registeredExpModules.values().stream().filter(m -> m instanceof CreatorModule)
				.map(CreatorModule.class::cast).collect(Collectors.toSet());
	}

	/** {@inheritDoc} */
	@Override
	public Set<OperatorModule<?>> getOperators() {
		return registeredExpModules.values().stream().filter(m -> m instanceof OperatorModule)
				.map(OperatorModule.class::cast).collect(Collectors.toSet());
	}

	/** {@inheritDoc} */
	@Override
	public Set<DecoderModule<?>> getDecoders() {
		return registeredExpModules.values().stream().filter(m -> m instanceof DecoderModule)
				.map(DecoderModule.class::cast).collect(Collectors.toSet());
	}

	/** {@inheritDoc} */
	@Override
	public Set<EvaluatorModule<?>> getEvaluators() {
		return registeredExpModules.values().stream().filter(m -> m instanceof EvaluatorModule)
				.map(EvaluatorModule.class::cast).collect(Collectors.toSet());
	}

	/** {@inheritDoc} */
	@Override
	public String getIntrospectionDescription() {
		// KEEP DESCRIPTION IN SYNC WITH THE SERVICE INTERFACE CLASS COMMENT!!!!
		return "Service for adding Exploration Modules to the DSE. The define the concrete problems"
				+ " that shall be solved by the DSE process. Modules registered with this service"
				+ " are known to the DSE and can be selected by users to define a DSE process.\n"
				+ "The Modules registered here are not defining a a concrete DSE process, just the"
				+ " concrete problems. Exploration modules can be categorized into three classes:"
				+ " Variable definitions, Problem statements and transformations, and Evaluators"
				+ " (solution metrics). The variable definitions consist of a Genotypes"
				+ " (variables), Operators (modify variables), and Creators"
				+ " (variable initialization). Problem statements transform Genotypes into"
				+ " evaluable solutions implemented by Decoders that produce"
				+ " Phenotypes. Evaluators rate Phenotypes such that the quality of a solution can"
				+ " be quantified.\n"
				+ "The complete DSE process is based on DI: each of the elements composing"
				+ " exploration modules have to be implemented as Guice AbstractModules.";
	}

	/** {@inheritDoc} */
	@Override
	public String getIntrospectionLabel() {
		return "Exploration Module Service";
	}

	/** {@inheritDoc} */
	@Override
	public boolean showInIntrospectionNavigation() {
		return true;
	}

	/** {@inheritDoc} */
	@Override
	public Collection<IIntrospectionItem> getIntrospectionItems() {
		return null;
	}

	/** {@inheritDoc} */
	@Override
	public IIntrospectionDetailsItem getDetailsItem() {
		return null;
	}
}
