/*-------------------------------------------------------------------------+
| Copyright 2016 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.util;

import static org.fortiss.af3.exploration.util.ExplorationDslModelElementFactory.createLocationPattern;
import static org.fortiss.af3.exploration.util.ExplorationDslModelElementFactory.createTwoProValCompExpression;

import org.fortiss.af3.component.model.Component;
import org.fortiss.af3.exploration.dsl.model.expressions.ConstraintExpression;
import org.fortiss.af3.exploration.dsl.model.operators.ArithmeticOperatorEnum;
import org.fortiss.af3.exploration.dsl.model.operators.ComparisonOperatorEnum;
import org.fortiss.af3.exploration.dsl.model.operators.LocationOperatorEnum;
import org.fortiss.af3.exploration.dsl.model.patterns.LocationPattern;
import org.fortiss.af3.exploration.dsl.model.patterns.SinglePropValueComparison;
import org.fortiss.af3.exploration.dsl.model.patterns.TwoPropertyValueComparison;
import org.fortiss.af3.exploration.model.ExplorationSpecification;
import org.fortiss.af3.exploration.model.ExplorationTarget;
import org.fortiss.af3.exploration.model.location.ComponentAllocationConstraint;
import org.fortiss.af3.exploration.model.location.ComponentDislocationConstraint;
import org.fortiss.af3.exploration.model.location.LocationConstraint;
import org.fortiss.af3.exploration.model.location.LocationFactory;
import org.fortiss.af3.exploration.model.time.DeadlineConstraint;
import org.fortiss.af3.exploration.model.time.PeriodConstraint;
import org.fortiss.af3.exploration.model.time.TimeFactory;
import org.fortiss.af3.platform.model.ExecutionUnit;
import org.fortiss.tooling.kernel.extension.data.ITopLevelElement;
import org.fortiss.tooling.kernel.service.IPersistencyService;

/**
 * Model element factory for {@link ExplorationTarget}s that are common for all DSE backends.
 * 
 * @author diewald
 */
public class CommonExplorationTagetModelElementFactory {

	// FIXME: Construct the Expression already if no parameters are passed & use setter methods in
	// the method of Allocation/DislocationConstraints.
	/** Creates a {@link ComponentAllocationConstraint}. */
	public static ComponentAllocationConstraint createComponentAllocationConstraint(Component comp,
			ExecutionUnit execUnit) {
		ComponentAllocationConstraint cAllocConstr = createComponentAllocationConstraint();
		// Construct the corresponding Expression via the LocationPattern.
		cAllocConstr.setMoeaExpression(createLocationPattern(comp, LocationOperatorEnum.ALLOCATION,
				execUnit));
		return cAllocConstr;
	}

	/** Creates a {@link ComponentAllocationConstraint}. */
	public static ComponentAllocationConstraint createComponentAllocationConstraint() {
		ComponentAllocationConstraint cAllocConstr =
				LocationFactory.eINSTANCE.createComponentAllocationConstraint();
		cAllocConstr.setName("Component Allocation Constraint");
		return cAllocConstr;
	}

	/** Creates a {@link ComponentDislocationConstraint}. */
	public static ComponentDislocationConstraint createComponentDislocationConstraint(
			Component comp, ExecutionUnit execUnit) {
		ComponentDislocationConstraint cDislocConstr = createComponentDislocationConstraint();
		// Construct the corresponding Expression via the LocationPattern.
		cDislocConstr.setMoeaExpression(createLocationPattern(comp,
				LocationOperatorEnum.DISLOCATION, execUnit));
		return cDislocConstr;
	}

	/** Creates a {@link ComponentDislocationConstraint}. */
	public static ComponentDislocationConstraint createComponentDislocationConstraint() {
		ComponentDislocationConstraint cDislocConstr =
				LocationFactory.eINSTANCE.createComponentDislocationConstraint();
		cDislocConstr.setName("Component Dislocation Constraint");
		return cDislocConstr;
	}

	/** Creates a deadline constraint within the goal specification of the current DSE model. */
	public static DeadlineConstraint addDeadlineConstraint(ExplorationSpecification targetSpec) {
		DeadlineConstraint deadlineConstraint = createDeadlineConstraint();
		ITopLevelElement modelContext =
				IPersistencyService.getInstance().getTopLevelElementFor(targetSpec);
		modelContext.runAsCommand(() -> targetSpec.getTargets().add(deadlineConstraint));
		return deadlineConstraint;
	}

	/** Creates an "empty" {@link DeadlineConstraint}. */
	public static DeadlineConstraint createDeadlineConstraint() {
		DeadlineConstraint deadlineConstr = TimeFactory.eINSTANCE.createDeadlineConstraint();
		deadlineConstr.setName("Deadline Constraint");
		TwoPropertyValueComparison<Double, Component, Component> twoPropValExpr =
				createTwoProValCompExpression(ArithmeticOperatorEnum.SUB,
						ComparisonOperatorEnum.LESS_OR_EQUAL);
		ConstraintExpression<TwoPropertyValueComparison<Double, Component, Component>> cExpr =
				ExplorationDslModelElementFactory.createConstraintExpression(null, null,
						twoPropValExpr);
		deadlineConstr.setMoeaExpression(cExpr);
		return deadlineConstr;
	}

	/** Creates an "empty" {@link PeriodConstraint}. */
	public static PeriodConstraint createPeriodConstraint() {
		PeriodConstraint periodConstr = TimeFactory.eINSTANCE.createPeriodConstraint();
		periodConstr.setName("Period Constraint");
		SinglePropValueComparison<Double, Component> singlePropValExpr =
				ExplorationDslModelElementFactory
						.createSinglePropValueCompExpression(ComparisonOperatorEnum.LESS_OR_EQUAL);
		ConstraintExpression<SinglePropValueComparison<Double, Component>> cExpr =
				ExplorationDslModelElementFactory.createConstraintExpression(null, null,
						singlePropValExpr);
		periodConstr.setMoeaExpression(cExpr);
		return periodConstr;
	}

	/** Creates a {@link LocationConstraint} by using a {@link LocationPattern}. */
	public static LocationConstraint<Component, ExecutionUnit> createLocationConstraint(
			LocationPattern<Component, ExecutionUnit> locationPattern) {
		LocationConstraint<Component, ExecutionUnit> createLocationConstraint =
				LocationFactory.eINSTANCE.createLocationConstraint();
		createLocationConstraint.setMoeaExpression(locationPattern);
		return createLocationConstraint;
	}
}
